"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var Type_1 = require("./Type");
var TypeNames_1 = require("./TypeNames");
function gatherNames(graph) {
    graph.allTypesUnordered().forEach(function (t) {
        if (t.hasNames) {
            t.getNames().clearInferred();
        }
    });
    var processed = immutable_1.Set();
    function processType(t, names, parentNames) {
        if (t.hasNames) {
            var alternatives_1 = [];
            names.forEach(function (name) {
                if (parentNames !== null) {
                    alternatives_1.push.apply(alternatives_1, parentNames.map(function (pn) { return pn + "_" + name; }).toArray());
                }
                alternatives_1.push(name + "_" + t.kind);
                if (parentNames !== null) {
                    alternatives_1.push.apply(alternatives_1, parentNames.map(function (pn) { return pn + "_" + name + "_" + t.kind; }).toArray());
                }
            });
            t.getNames().add(new TypeNames_1.TypeNames(names, immutable_1.OrderedSet(alternatives_1), true));
        }
        var processedEnry = immutable_1.List([t, names, parentNames]);
        if (processed.has(processedEnry))
            return;
        processed = processed.add(processedEnry);
        Type_1.matchCompoundType(t, function (arrayType) {
            processType(arrayType.items, names.map(pluralize.singular), parentNames);
        }, function (classType) {
            var properties = classType.properties.sortBy(function (_, n) { return n; });
            properties.forEach(function (propertyType, propertyName) {
                processType(propertyType, immutable_1.OrderedSet([propertyName]), names);
            });
        }, function (mapType) {
            processType(mapType.values, names.map(pluralize.singular), parentNames);
        }, function (unionType) {
            var members = unionType.members.sortBy(function (member) { return member.kind; });
            members.forEach(function (memberType) {
                processType(memberType, names, parentNames);
            });
        });
    }
    graph.topLevels.forEach(function (t, name) {
        processType(t, immutable_1.OrderedSet([name]), null);
    });
}
exports.gatherNames = gatherNames;
