"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var MIN_LENGTH_FOR_ENUM = 10;
function shouldBeEnum(t) {
    var enumCases = t.enumCases;
    if (enumCases !== undefined) {
        Support_1.assert(enumCases.size > 0, "How did we end up with zero enum cases?");
        var numValues = enumCases.map(function (n) { return n; }).reduce(function (a, b) { return a + b; });
        if (numValues >= MIN_LENGTH_FOR_ENUM && enumCases.size < Math.sqrt(numValues)) {
            return t.enumCases;
        }
    }
    return undefined;
}
function replaceString(t, builder) {
    var maybeEnumCases = shouldBeEnum(t);
    if (maybeEnumCases !== undefined) {
        return builder.getEnumType(t.getNames(), maybeEnumCases.keySeq().toOrderedSet());
    }
    return builder.getStringType(undefined, undefined);
}
function unionNeedsReplacing(u) {
    var stringMembers = u.stringTypeMembers;
    if (stringMembers.size <= 1)
        return undefined;
    var stringType = u.findMember("string");
    if (stringType === undefined || shouldBeEnum(stringType) !== undefined)
        return undefined;
    return stringMembers;
}
function replaceUnion(u, builder) {
    var stringMembers = Support_1.defined(unionNeedsReplacing(u));
    var types = [];
    u.members.forEach(function (t) {
        if (stringMembers.has(t))
            return;
        types.push(builder.reconstituteType(t));
    });
    // FIXME: add names
    var stringType = builder.getStringType(undefined, undefined);
    if (types.length === 0) {
        return stringType;
    }
    types.push(stringType);
    return builder.getUnionType(u.getNames(), immutable_1.OrderedSet(types));
}
function replace(setOfStringOrUnion, builder) {
    Support_1.assert(setOfStringOrUnion.size === 1);
    var t = Support_1.defined(setOfStringOrUnion.first());
    if (t instanceof Type_1.StringType) {
        return replaceString(t, builder);
    }
    else if (t instanceof Type_1.UnionType) {
        return replaceUnion(t, builder);
    }
    else {
        return Support_1.assertNever(t);
    }
}
function inferEnums(graph, stringTypeMapping) {
    var allStrings = graph
        .allTypesUnordered()
        .filter(function (t) { return t instanceof Type_1.StringType; })
        .map(function (t) { return [t]; })
        .toArray();
    var allUnions = graph.allNamedTypesSeparated().unions;
    var unionsToReplace = allUnions
        .filter(unionNeedsReplacing)
        .map(function (t) { return [t]; })
        .toArray();
    var typesToReplace = [].concat(allStrings, unionsToReplace);
    return graph.rewrite(stringTypeMapping, typesToReplace, replace);
}
exports.inferEnums = inferEnums;
