"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("../Type");
var Source_1 = require("../Source");
var Strings_1 = require("../Strings");
var Support_1 = require("../Support");
var Naming_1 = require("../Naming");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var TargetLanguage_1 = require("../TargetLanguage");
var RendererOptions_1 = require("../RendererOptions");
var Annotation_1 = require("../Annotation");
var unicode = require("unicode-properties");
var lodash = require("lodash");
var CSharpTargetLanguage = /** @class */ (function (_super) {
    __extends(CSharpTargetLanguage, _super);
    function CSharpTargetLanguage() {
        var _this = _super.call(this, "C#", ["cs", "csharp"], "cs") || this;
        _this._listOption = new RendererOptions_1.EnumOption("array-type", "Use T[] or List<T>", [
            ["array", false],
            ["list", true]
        ]);
        _this._denseOption = new RendererOptions_1.EnumOption("density", "Property density", [["normal", false], ["dense", true]]);
        _this._featuresOption = new RendererOptions_1.EnumOption("features", "Output features", [
            ["complete", { helpers: true, attributes: true }],
            ["attributes-only", { helpers: false, attributes: true }],
            ["just-types", { helpers: false, attributes: false }]
        ]);
        // FIXME: Do this via a configurable named eventually.
        _this._namespaceOption = new RendererOptions_1.StringOption("namespace", "Generated namespace", "NAME", "QuickType");
        _this._versionOption = new RendererOptions_1.EnumOption("csharp-version", "C# version", [["6", 6], ["5", 5]]);
        _this.setOptions([
            _this._namespaceOption,
            _this._versionOption,
            _this._denseOption,
            _this._listOption,
            _this._featuresOption
        ]);
        return _this;
    }
    Object.defineProperty(CSharpTargetLanguage.prototype, "partialStringTypeMapping", {
        get: function () {
            return { date: "date-time", time: "date-time", dateTime: "date-time" };
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CSharpTargetLanguage.prototype, "rendererClass", {
        get: function () {
            return CSharpRenderer;
        },
        enumerable: true,
        configurable: true
    });
    return CSharpTargetLanguage;
}(TargetLanguage_1.TargetLanguage));
exports.default = CSharpTargetLanguage;
var namingFunction = Naming_1.funPrefixNamer(csNameStyle);
// FIXME: Make a Named?
var denseJsonPropertyName = "J";
function isStartCharacter(utf16Unit) {
    if (unicode.isAlphabetic(utf16Unit)) {
        return true;
    }
    return utf16Unit === 0x5f; // underscore
}
function isPartCharacter(utf16Unit) {
    var category = unicode.getCategory(utf16Unit);
    if (["Nd", "Pc", "Mn", "Mc"].indexOf(category) >= 0) {
        return true;
    }
    return isStartCharacter(utf16Unit);
}
var legalizeName = Strings_1.utf16LegalizeCharacters(isPartCharacter);
function csNameStyle(original) {
    var words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, "", isStartCharacter);
}
function isValueType(t) {
    return ["integer", "double", "bool", "enum", "date-time"].indexOf(t.kind) >= 0;
}
var CSharpRenderer = /** @class */ (function (_super) {
    __extends(CSharpRenderer, _super);
    function CSharpRenderer(graph, leadingComments, _namespaceName, _version, _dense, _useList, outputFeatures) {
        var _this = _super.call(this, graph, leadingComments) || this;
        _this._namespaceName = _namespaceName;
        _this._version = _version;
        _this._dense = _dense;
        _this._useList = _useList;
        _this._enumExtensionsNames = immutable_1.Map();
        _this.emitBlock = function (f, semicolon) {
            if (semicolon === void 0) { semicolon = false; }
            _this.emitLine("{");
            _this.indent(f);
            _this.emitLine("}", semicolon ? ";" : "");
        };
        _this.csType = function (t, withIssues) {
            if (withIssues === void 0) { withIssues = false; }
            return Type_1.matchType(t, function (_anyType) { return Source_1.maybeAnnotated(withIssues, Annotation_1.anyTypeIssueAnnotation, "object"); }, function (_nullType) { return Source_1.maybeAnnotated(withIssues, Annotation_1.nullTypeIssueAnnotation, "object"); }, function (_boolType) { return "bool"; }, function (_integerType) { return "long"; }, function (_doubleType) { return "double"; }, function (_stringType) { return "string"; }, function (arrayType) {
                var itemsType = _this.csType(arrayType.items, withIssues);
                if (_this._useList) {
                    return ["List<", itemsType, ">"];
                }
                else {
                    return [itemsType, "[]"];
                }
            }, function (classType) { return _this.nameForNamedType(classType); }, function (mapType) { return ["Dictionary<string, ", _this.csType(mapType.values, withIssues), ">"]; }, function (enumType) { return _this.nameForNamedType(enumType); }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable)
                    return _this.nullableCSType(nullable);
                return _this.nameForNamedType(unionType);
            }, {
                dateTimeType: function (_) { return "DateTime"; }
            });
        };
        _this.nullableCSType = function (t) {
            var csType = _this.csType(t);
            if (isValueType(t)) {
                return [csType, "?"];
            }
            else {
                return csType;
            }
        };
        _this.emitClass = function (isPublic, declaration, name, emitter) {
            if (isPublic) {
                declaration = ["public ", declaration];
            }
            _this.emitLine(declaration, " ", name);
            _this.emitBlock(emitter);
        };
        _this.emitClassDefinition = function (c, className) {
            var jsonProperty = _this._dense ? denseJsonPropertyName : "JsonProperty";
            _this.emitClass(true, [_this.partialString, "class"], className, function () {
                if (c.properties.isEmpty())
                    return;
                var maxWidth = Support_1.defined(c.properties.map(function (_, name) { return Strings_1.utf16StringEscape(name).length; }).max());
                var blankLines = _this._needAttributes && !_this._dense ? "interposing" : "none";
                _this.forEachClassProperty(c, blankLines, function (name, jsonName, t) {
                    var csType = _this.csType(t, true);
                    var escapedName = Strings_1.utf16StringEscape(jsonName);
                    var attribute = ["[", jsonProperty, '("', escapedName, '")]'];
                    var property = ["public ", csType, " ", name, " { get; set; }"];
                    if (!_this._needAttributes) {
                        _this.emitLine(property);
                    }
                    else if (_this._dense) {
                        var indent = maxWidth - escapedName.length + 1;
                        var whitespace = lodash.repeat(" ", indent);
                        _this.emitLine(attribute, whitespace, property);
                    }
                    else {
                        _this.emitLine(attribute);
                        _this.emitLine(property);
                    }
                });
            });
        };
        _this.emitUnionDefinition = function (u, unionName) {
            var nonNulls = Type_1.removeNullFromUnion(u)[1];
            _this.emitClass(true, [_this.partialString, "struct"], unionName, function () {
                _this.forEachUnionMember(u, nonNulls, "none", null, function (fieldName, t) {
                    var csType = _this.nullableCSType(t);
                    _this.emitLine("public ", csType, " ", fieldName, ";");
                });
            });
        };
        _this.emitEnumDefinition = function (e, enumName) {
            var caseNames = [];
            _this.forEachEnumCase(e, "none", function (name) {
                if (caseNames.length > 0)
                    caseNames.push(", ");
                caseNames.push(name);
            });
            _this.emitLine("public enum ", enumName, " { ", caseNames, " };");
        };
        _this.emitFromJsonForTopLevel = function (t, name) {
            var partial;
            var typeKind;
            var definedType = _this.namedTypeToNameForTopLevel(t);
            if (definedType) {
                partial = "partial ";
                typeKind = definedType instanceof Type_1.ClassType ? "class" : "struct";
            }
            else {
                partial = "";
                typeKind = "class";
            }
            var csType = _this.csType(t);
            _this.emitClass(true, [partial, typeKind], name, function () {
                // FIXME: Make FromJson a Named
                _this.emitExpressionMember(["public static ", csType, " FromJson(string json)"], ["JsonConvert.DeserializeObject<", csType, ">(json, Converter.Settings)"]);
            });
        };
        _this.emitEnumExtension = function (e, enumName) {
            _this.emitClass(false, "static class", Support_1.defined(_this._enumExtensionsNames.get(enumName)), function () {
                _this.emitLine("public static ", enumName, "? ValueForString(string str)");
                _this.emitBlock(function () {
                    _this.emitLine("switch (str)");
                    _this.emitBlock(function () {
                        _this.forEachEnumCase(e, "none", function (name, jsonName) {
                            _this.emitLine('case "', Strings_1.utf16StringEscape(jsonName), '": return ', enumName, ".", name, ";");
                        });
                        _this.emitLine("default: return null;");
                    });
                });
                _this.ensureBlankLine();
                _this.emitLine("public static ", enumName, " ReadJson(JsonReader reader, JsonSerializer serializer)");
                _this.emitBlock(function () {
                    _this.emitLine("var str = serializer.Deserialize<string>(reader);");
                    _this.emitLine("var maybeValue = ValueForString(str);");
                    _this.emitLine("if (maybeValue.HasValue) return maybeValue.Value;");
                    _this.emitLine('throw new Exception("Unknown enum case " + str);');
                });
                _this.ensureBlankLine();
                _this.emitLine("public static void WriteJson(this ", enumName, " value, JsonWriter writer, JsonSerializer serializer)");
                _this.emitBlock(function () {
                    _this.emitLine("switch (value)");
                    _this.emitBlock(function () {
                        _this.forEachEnumCase(e, "none", function (name, jsonName) {
                            _this.emitLine("case ", enumName, ".", name, ': serializer.Serialize(writer, "', Strings_1.utf16StringEscape(jsonName), '"); break;');
                        });
                    });
                });
            });
        };
        _this.emitUnionJSONPartial = function (u, unionName) {
            var tokenCase = function (tokenType) {
                _this.emitLine("case JsonToken.", tokenType, ":");
            };
            var emitNullDeserializer = function () {
                tokenCase("Null");
                _this.indent(function () { return _this.emitLine("return;"); });
            };
            var emitDeserializeType = function (t) {
                _this.emitLine(_this.nameForUnionMember(u, t), " = serializer.Deserialize<", _this.csType(t), ">(reader);");
                _this.emitLine("return;");
            };
            var emitDeserializer = function (tokenType, kind) {
                var t = u.findMember(kind);
                if (!t)
                    return;
                tokenCase(tokenType);
                _this.indent(function () { return emitDeserializeType(t); });
            };
            var emitDoubleDeserializer = function () {
                var t = u.findMember("double");
                if (!t)
                    return;
                if (!u.findMember("integer"))
                    tokenCase("Integer");
                tokenCase("Float");
                _this.indent(function () { return emitDeserializeType(t); });
            };
            var emitStringDeserializer = function () {
                var stringTypes = u.stringTypeMembers;
                if (stringTypes.isEmpty())
                    return;
                tokenCase("String");
                tokenCase("Date");
                _this.indent(function () {
                    if (stringTypes.size === 1) {
                        emitDeserializeType(Support_1.defined(stringTypes.first()));
                        return;
                    }
                    _this.emitLine("var str = serializer.Deserialize<string>(reader);");
                    _this.forEachUnionMember(u, stringTypes, "none", null, function (fieldName, t) {
                        if (t instanceof Type_1.EnumType) {
                            var extension = Support_1.defined(_this._enumExtensionsNames.get(_this.nameForNamedType(t)));
                            _this.emitLine("var maybeEnum = ", extension, ".ValueForString(str);");
                            _this.emitLine("if (maybeEnum.HasValue)");
                            _this.emitBlock(function () {
                                _this.emitLine(fieldName, " = maybeEnum.Value;");
                                _this.emitLine("return;");
                            });
                        }
                        else if (t.kind === "date-time") {
                            _this.emitLine("DateTime dt;");
                            _this.emitLine("if (System.DateTime.TryParse(str, out dt))");
                            _this.emitBlock(function () {
                                _this.emitLine(fieldName, " = dt;");
                                _this.emitLine("return;");
                            });
                        }
                        else {
                            return Support_1.panic("Unsupported string enum type " + t.kind);
                        }
                    });
                    _this.emitLine("break;");
                });
            };
            var _a = Type_1.removeNullFromUnion(u), hasNull = _a[0], nonNulls = _a[1];
            _this.emitClass(true, "partial struct", unionName, function () {
                _this.emitLine("public ", unionName, "(JsonReader reader, JsonSerializer serializer)");
                _this.emitBlock(function () {
                    _this.forEachUnionMember(u, nonNulls, "none", null, function (fieldName, _) {
                        _this.emitLine(fieldName, " = null;");
                    });
                    _this.ensureBlankLine();
                    _this.emitLine("switch (reader.TokenType)");
                    _this.emitBlock(function () {
                        if (hasNull)
                            emitNullDeserializer();
                        emitDeserializer("Integer", "integer");
                        emitDoubleDeserializer();
                        emitDeserializer("Boolean", "bool");
                        emitDeserializer("StartArray", "array");
                        emitDeserializer("StartObject", "class");
                        emitDeserializer("StartObject", "map");
                        emitStringDeserializer();
                    });
                    _this.emitLine('throw new Exception("Cannot convert ', unionName, '");');
                });
                _this.ensureBlankLine();
                _this.emitLine("public void WriteJson(JsonWriter writer, JsonSerializer serializer)");
                _this.emitBlock(function () {
                    _this.forEachUnionMember(u, nonNulls, "none", null, function (fieldName, _) {
                        _this.emitLine("if (", fieldName, " != null)");
                        _this.emitBlock(function () {
                            _this.emitLine("serializer.Serialize(writer, ", fieldName, ");");
                            _this.emitLine("return;");
                        });
                    });
                    if (hasNull) {
                        _this.emitLine("writer.WriteNull();");
                    }
                    else {
                        _this.emitLine('throw new Exception("Union must not be null");');
                    }
                });
            });
        };
        _this.emitSerializeClass = function () {
            // FIXME: Make Serialize a Named
            _this.emitClass(true, "static class", "Serialize", function () {
                _this.topLevels.forEach(function (t, _) {
                    // FIXME: Make ToJson a Named
                    _this.emitExpressionMember(["public static string ToJson(this ", _this.csType(t), " self)"], "JsonConvert.SerializeObject(self, Converter.Settings)");
                });
            });
        };
        _this.emitConverterMembers = function () {
            var enumNames = _this.enums.map(_this.nameForNamedType);
            var unionNames = _this.namedUnions.map(_this.nameForNamedType);
            var allNames = enumNames.union(unionNames);
            var canConvertExprs = allNames.map(function (n) { return ["t == typeof(", n, ")"]; });
            var nullableCanConvertExprs = enumNames.map(function (n) { return ["t == typeof(", n, "?)"]; });
            var canConvertExpr = Support_1.intercalate(" || ", canConvertExprs.union(nullableCanConvertExprs));
            // FIXME: make Iterable<any, Sourcelike> a Sourcelike, too?
            _this.emitExpressionMember("public override bool CanConvert(Type t)", canConvertExpr.toArray());
            _this.ensureBlankLine();
            _this.emitLine("public override object ReadJson(JsonReader reader, Type t, object existingValue, JsonSerializer serializer)");
            _this.emitBlock(function () {
                _this.emitTypeSwitch(enumNames, function (t) { return ["t == typeof(", t, ")"]; }, false, false, function (n) {
                    var extensionsName = Support_1.defined(_this._enumExtensionsNames.get(n));
                    _this.emitLine("return ", extensionsName, ".ReadJson(reader, serializer);");
                });
                _this.emitTypeSwitch(enumNames, function (t) { return ["t == typeof(", t, "?)"]; }, true, false, function (n) {
                    _this.emitLine("if (reader.TokenType == JsonToken.Null) return null;");
                    var extensionsName = Support_1.defined(_this._enumExtensionsNames.get(n));
                    _this.emitLine("return ", extensionsName, ".ReadJson(reader, serializer);");
                });
                // FIXME: call the constructor via reflection?
                _this.emitTypeSwitch(unionNames, function (t) { return ["t == typeof(", t, ")"]; }, false, false, function (n) {
                    _this.emitLine("return new ", n, "(reader, serializer);");
                });
                _this.emitLine('throw new Exception("Unknown type");');
            });
            _this.ensureBlankLine();
            _this.emitLine("public override void WriteJson(JsonWriter writer, object value, JsonSerializer serializer)");
            _this.emitBlock(function () {
                _this.emitLine("var t = value.GetType();");
                _this.emitTypeSwitch(allNames, function (t) { return ["t == typeof(", t, ")"]; }, true, true, function (n) {
                    _this.emitLine("((", n, ")value).WriteJson(writer, serializer);");
                });
                _this.emitLine('throw new Exception("Unknown type");');
            });
        };
        _this.emitConverterClass = function () {
            var jsonConverter = _this.haveEnums || _this.haveNamedUnions;
            // FIXME: Make Converter a Named
            var converterName = ["Converter"];
            if (jsonConverter)
                converterName = converterName.concat([": JsonConverter"]);
            _this.emitClass(true, "class", converterName, function () {
                if (jsonConverter) {
                    _this.emitConverterMembers();
                    _this.ensureBlankLine();
                }
                _this.emitLine("public static readonly JsonSerializerSettings Settings = new JsonSerializerSettings");
                _this.emitBlock(function () {
                    _this.emitLine("MetadataPropertyHandling = MetadataPropertyHandling.Ignore,");
                    _this.emitLine("DateParseHandling = DateParseHandling.None,");
                    if (_this.haveNamedUnions || _this.haveEnums) {
                        _this.emitLine("Converters = { new Converter() },");
                    }
                }, true);
            });
        };
        _this.emitTypesAndSupport = function () {
            var using = function (ns) {
                _this.emitLine("using ", ns, ";");
            };
            if (_this._needAttributes || _this._needHelpers) {
                for (var _i = 0, _a = ["System", "System.Net", "System.Collections.Generic"]; _i < _a.length; _i++) {
                    var ns = _a[_i];
                    using(ns);
                }
                _this.ensureBlankLine();
                using("Newtonsoft.Json");
                if (_this._dense) {
                    using([denseJsonPropertyName, " = Newtonsoft.Json.JsonPropertyAttribute"]);
                }
            }
            _this.forEachClass("leading-and-interposing", _this.emitClassDefinition);
            _this.forEachEnum("leading-and-interposing", _this.emitEnumDefinition);
            _this.forEachUnion("leading-and-interposing", _this.emitUnionDefinition);
            if (_this._needHelpers) {
                _this.forEachTopLevel("leading-and-interposing", _this.emitFromJsonForTopLevel);
                _this.forEachEnum("leading-and-interposing", _this.emitEnumExtension);
                _this.forEachUnion("leading-and-interposing", _this.emitUnionJSONPartial);
                _this.ensureBlankLine();
                _this.emitSerializeClass();
            }
            if (_this._needHelpers || (_this._needAttributes && (_this.haveNamedUnions || _this.haveEnums))) {
                _this.ensureBlankLine();
                _this.emitConverterClass();
            }
        };
        _this._needHelpers = outputFeatures.helpers;
        _this._needAttributes = outputFeatures.attributes;
        return _this;
    }
    Object.defineProperty(CSharpRenderer.prototype, "forbiddenNamesForGlobalNamespace", {
        get: function () {
            return ["QuickType", "Converter", "JsonConverter", "Type", "Serialize"];
        },
        enumerable: true,
        configurable: true
    });
    CSharpRenderer.prototype.forbiddenForClassProperties = function (_, classNamed) {
        return { names: [classNamed], namespaces: [] };
    };
    CSharpRenderer.prototype.topLevelNameStyle = function (rawName) {
        return csNameStyle(rawName);
    };
    Object.defineProperty(CSharpRenderer.prototype, "namedTypeNamer", {
        get: function () {
            return namingFunction;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CSharpRenderer.prototype, "classPropertyNamer", {
        get: function () {
            return namingFunction;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CSharpRenderer.prototype, "unionMemberNamer", {
        get: function () {
            return namingFunction;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CSharpRenderer.prototype, "enumCaseNamer", {
        get: function () {
            return namingFunction;
        },
        enumerable: true,
        configurable: true
    });
    CSharpRenderer.prototype.unionNeedsName = function (u) {
        return !Type_1.nullableFromUnion(u);
    };
    CSharpRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        var definedTypes = type.directlyReachableTypes(function (t) {
            if ((!(t instanceof Type_1.UnionType) && Type_1.isNamedType(t)) || (t instanceof Type_1.UnionType && !Type_1.nullableFromUnion(t))) {
                return immutable_1.OrderedSet([t]);
            }
            return null;
        });
        Support_1.assert(definedTypes.size <= 1, "Cannot have more than one defined type per top-level");
        // If the top-level type doesn't contain any classes or unions
        // we have to define a class just for the `FromJson` method, in
        // emitFromJsonForTopLevel.
        var first = definedTypes.first();
        if (first === undefined)
            return null;
        return first;
    };
    CSharpRenderer.prototype.namedTypeDependencyNames = function (t, name) {
        if (!(t instanceof Type_1.EnumType))
            return [];
        var extensionsName = new Naming_1.DependencyName(namingFunction, function (lookup) { return lookup(name) + "_extensions"; });
        this._enumExtensionsNames = this._enumExtensionsNames.set(name, extensionsName);
        return [extensionsName];
    };
    Object.defineProperty(CSharpRenderer.prototype, "partialString", {
        get: function () {
            return this._needHelpers ? "partial " : "";
        },
        enumerable: true,
        configurable: true
    });
    CSharpRenderer.prototype.emitExpressionMember = function (declare, define) {
        var _this = this;
        if (this._version === 5) {
            this.emitLine(declare);
            this.emitBlock(function () {
                _this.emitLine("return ", define, ";");
            });
        }
        else {
            this.emitLine(declare, " => ", define, ";");
        }
    };
    CSharpRenderer.prototype.emitTypeSwitch = function (types, condition, withBlock, withReturn, f) {
        var _this = this;
        Support_1.assert(!withReturn || withBlock, "Can only have return with block");
        types.forEach(function (t) {
            _this.emitLine("if (", condition(t), ")");
            if (withBlock) {
                _this.emitBlock(function () {
                    f(t);
                    if (withReturn) {
                        _this.emitLine("return;");
                    }
                });
            }
            else {
                _this.indent(function () { return f(t); });
            }
        });
    };
    CSharpRenderer.prototype.emitSourceStructure = function () {
        var _this = this;
        if (this.leadingComments !== undefined) {
            this.emitCommentLines("// ", this.leadingComments);
        }
        else if (this._needHelpers) {
            this.emitLine("// To parse this JSON data, add NuGet 'Newtonsoft.Json' then do", this.topLevels.size === 1 ? "" : " one of these", ":");
            this.emitLine("//");
            this.emitLine("//    using ", this._namespaceName, ";");
            this.forEachTopLevel("none", function (_, topLevelName) {
                _this.emitLine("//");
                _this.emitLine("//    var data = ", topLevelName, ".FromJson(jsonString);");
            });
        }
        this.ensureBlankLine();
        if (this._needHelpers || this._needAttributes) {
            this.emitLine("namespace ", this._namespaceName);
            this.emitBlock(this.emitTypesAndSupport);
        }
        else {
            this.emitTypesAndSupport();
        }
    };
    return CSharpRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
