"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var TargetLanguage_1 = require("../TargetLanguage");
var Type_1 = require("../Type");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var Naming_1 = require("../Naming");
var Strings_1 = require("../Strings");
var Support_1 = require("../Support");
var JSONSchemaTargetLanguage = /** @class */ (function (_super) {
    __extends(JSONSchemaTargetLanguage, _super);
    function JSONSchemaTargetLanguage() {
        var _this = _super.call(this, "JSON Schema", ["schema", "json-schema"], "schema") || this;
        _this.setOptions([]);
        return _this;
    }
    Object.defineProperty(JSONSchemaTargetLanguage.prototype, "partialStringTypeMapping", {
        get: function () {
            return { date: "date", time: "time", dateTime: "date-time" };
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(JSONSchemaTargetLanguage.prototype, "rendererClass", {
        get: function () {
            return JSONSchemaRenderer;
        },
        enumerable: true,
        configurable: true
    });
    return JSONSchemaTargetLanguage;
}(TargetLanguage_1.TargetLanguage));
exports.default = JSONSchemaTargetLanguage;
var namingFunction = Naming_1.funPrefixNamer(jsonNameStyle);
var legalizeName = Strings_1.legalizeCharacters(function (cp) { return cp >= 32 && cp < 128 && cp !== 0x2f; } /* slash */);
function jsonNameStyle(original) {
    var words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.allUpperWordStyle, Strings_1.allUpperWordStyle, "", function (_) { return true; });
}
var JSONSchemaRenderer = /** @class */ (function (_super) {
    __extends(JSONSchemaRenderer, _super);
    function JSONSchemaRenderer() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.nameForType = function (t) {
            return Support_1.defined(_this.names.get(_this.nameForNamedType(t)));
        };
        _this.makeOneOf = function (types) {
            var count = types.count();
            Support_1.assert(count > 0, "Must have at least one type for oneOf");
            if (count === 1) {
                return _this.schemaForType(Support_1.defined(types.first()));
            }
            return { oneOf: types.map(_this.schemaForType).toArray() };
        };
        _this.schemaForType = function (t) {
            return Type_1.matchTypeExhaustive(t, function (_anyType) { return ({}); }, function (_nullType) { return ({ type: "null" }); }, function (_boolType) { return ({ type: "boolean" }); }, function (_integerType) { return ({ type: "integer" }); }, function (_doubleType) { return ({ type: "number" }); }, function (_stringType) { return ({ type: "string" }); }, function (arrayType) { return ({ type: "array", items: _this.schemaForType(arrayType.items) }); }, function (classType) { return ({ $ref: "#/definitions/" + _this.nameForType(classType) }); }, function (mapType) { return ({ type: "object", additionalProperties: _this.schemaForType(mapType.values) }); }, function (enumType) { return ({ type: "string", enum: enumType.cases.toArray(), title: enumType.getCombinedName() }); }, function (unionType) {
                var schema = _this.makeOneOf(unionType.sortedMembers);
                schema.title = unionType.getCombinedName();
                return schema;
            }, function (_dateType) { return ({ type: "string", format: "date" }); }, function (_timeType) { return ({ type: "string", format: "time" }); }, function (_dateTimeType) { return ({ type: "string", format: "date-time" }); });
        };
        _this.definitionForClass = function (c) {
            var properties = {};
            var required = [];
            c.properties.forEach(function (t, name) {
                properties[name] = _this.schemaForType(t);
                if (!t.isNullable) {
                    required.push(name);
                }
            });
            return {
                type: "object",
                additionalProperties: false,
                properties: properties,
                required: required.sort(),
                title: c.getCombinedName()
            };
        };
        return _this;
    }
    JSONSchemaRenderer.prototype.topLevelNameStyle = function (rawName) {
        return jsonNameStyle(rawName);
    };
    Object.defineProperty(JSONSchemaRenderer.prototype, "namedTypeNamer", {
        get: function () {
            return namingFunction;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(JSONSchemaRenderer.prototype, "classPropertyNamer", {
        get: function () {
            return null;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(JSONSchemaRenderer.prototype, "unionMemberNamer", {
        get: function () {
            return null;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(JSONSchemaRenderer.prototype, "enumCaseNamer", {
        get: function () {
            return null;
        },
        enumerable: true,
        configurable: true
    });
    JSONSchemaRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        if (Type_1.isNamedType(type)) {
            return type;
        }
        return null;
    };
    JSONSchemaRenderer.prototype.unionNeedsName = function (_) {
        return false;
    };
    JSONSchemaRenderer.prototype.emitSourceStructure = function () {
        var _this = this;
        // FIXME: Find a better way to do multiple top-levels.  Maybe multiple files?
        var schema = this.makeOneOf(this.topLevels);
        var definitions = {};
        this.forEachClass("none", function (c, name) {
            definitions[Support_1.defined(_this.names.get(name))] = _this.definitionForClass(c);
        });
        schema.definitions = definitions;
        this.emitMultiline(JSON.stringify(schema, undefined, "    "));
    };
    return JSONSchemaRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
