"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var Support_1 = require("./Support");
function makeTypeNames(nameOrNames, areNamesInferred) {
    if (typeof nameOrNames === "string") {
        return new TypeNames(immutable_1.OrderedSet([nameOrNames]), immutable_1.OrderedSet(), areNamesInferred);
    }
    else {
        return nameOrNames;
    }
}
exports.makeTypeNames = makeTypeNames;
// FIXME: In the case of overlapping prefixes and suffixes we will
// produce a name that includes the overlap twice.  For example, for
// the names "aaa" and "aaaa" we have the common prefix "aaa" and the
// common suffix "aaa", so we will produce the combined name "aaaaaa".
function combineNames(names) {
    var first = names.first();
    if (first === undefined) {
        return Support_1.panic("Named type has no names");
    }
    if (names.count() === 1) {
        return first;
    }
    var prefixLength = first.length;
    var suffixLength = first.length;
    names.rest().forEach(function (n) {
        prefixLength = Math.min(prefixLength, n.length);
        for (var i = 0; i < prefixLength; i++) {
            if (first[i] !== n[i]) {
                prefixLength = i;
                break;
            }
        }
        suffixLength = Math.min(suffixLength, n.length);
        for (var i = 0; i < suffixLength; i++) {
            if (first[first.length - i - 1] !== n[n.length - i - 1]) {
                suffixLength = i;
                break;
            }
        }
    });
    var prefix = prefixLength > 2 ? first.substr(0, prefixLength) : "";
    var suffix = suffixLength > 2 ? first.substr(first.length - suffixLength) : "";
    var combined = prefix + suffix;
    if (combined.length > 2) {
        return combined;
    }
    return first;
}
var TypeNames = /** @class */ (function () {
    function TypeNames(_names, _alternativeNames, _areInferred) {
        var _this = this;
        this._names = _names;
        this._alternativeNames = _alternativeNames;
        this._areInferred = _areInferred;
        this.copy = function () {
            return new TypeNames(_this._names, _this._alternativeNames, _this._areInferred);
        };
        this.clearInferred = function () {
            if (_this._areInferred) {
                _this._names = immutable_1.OrderedSet();
            }
            _this._alternativeNames = immutable_1.OrderedSet();
        };
        this.makeInferred = function () {
            if (_this._areInferred)
                return _this;
            return new TypeNames(_this._names, _this._alternativeNames, true);
        };
        this.singularize = function () {
            return new TypeNames(_this._names.map(pluralize.singular), _this._alternativeNames.map(pluralize.singular), true);
        };
    }
    Object.defineProperty(TypeNames.prototype, "names", {
        get: function () {
            return this._names;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TypeNames.prototype, "areInferred", {
        get: function () {
            return this._areInferred;
        },
        enumerable: true,
        configurable: true
    });
    TypeNames.prototype.add = function (names) {
        if (this._areInferred && !names._areInferred) {
            this._names = names._names;
            this._areInferred = false;
        }
        else if (this._areInferred === names._areInferred) {
            this._names = this._names.union(names._names);
        }
        this._alternativeNames = this._alternativeNames.union(names._alternativeNames);
    };
    TypeNames.prototype.union = function (names) {
        var copy = this.copy();
        copy.add(names);
        return copy;
    };
    Object.defineProperty(TypeNames.prototype, "combinedName", {
        get: function () {
            return combineNames(this._names);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TypeNames.prototype, "proposedNames", {
        get: function () {
            return immutable_1.OrderedSet([this.combinedName]).union(this._alternativeNames);
        },
        enumerable: true,
        configurable: true
    });
    return TypeNames;
}());
exports.TypeNames = TypeNames;
function typeNamesUnion(c) {
    var names = new TypeNames(immutable_1.OrderedSet(), immutable_1.OrderedSet(), true);
    c.forEach(function (n) { return names.add(n); });
    return names;
}
exports.typeNamesUnion = typeNamesUnion;
