"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var TypeBuilder_1 = require("./TypeBuilder");
var TypeNames_1 = require("./TypeNames");
var Support_1 = require("./Support");
function getCliqueProperties(clique, makePropertyType) {
    var properties = immutable_1.OrderedMap();
    for (var _i = 0, clique_1 = clique; _i < clique_1.length; _i++) {
        var c = clique_1[_i];
        c.properties.forEach(function (t, name) {
            var p = properties.get(name);
            if (p === undefined) {
                p = [immutable_1.OrderedSet(), 0, false];
                properties = properties.set(name, p);
            }
            p[1] += 1;
            p[0] = p[0].union(Type_1.nonNullTypeCases(t));
            if (t.isNullable) {
                p[2] = true;
            }
        });
    }
    return properties.map(function (_a, name) {
        var types = _a[0], count = _a[1], haveNullable = _a[2];
        Support_1.assert(!types.some(function (t) { return t.isNullable || t.kind === "any"; }), "Nullable types are not allowed in non-nullable properties");
        var isNullable = haveNullable || count < clique.length;
        var allNames = types.filter(function (t) { return t.hasNames; }).map(function (t) { return t.getNames(); });
        var typeNames = allNames.isEmpty() ? TypeNames_1.makeTypeNames(name, true) : TypeNames_1.typeNamesUnion(allNames);
        return makePropertyType(typeNames, types, isNullable);
    });
}
var UnifyUnionBuilder = /** @class */ (function (_super) {
    __extends(UnifyUnionBuilder, _super);
    function UnifyUnionBuilder(typeBuilder, typeNames, _makeEnums, _unifyTypes) {
        var _this = _super.call(this, typeBuilder, typeNames) || this;
        _this._makeEnums = _makeEnums;
        _this._unifyTypes = _unifyTypes;
        return _this;
    }
    UnifyUnionBuilder.prototype.makeEnum = function (enumCases, counts) {
        if (this._makeEnums) {
            return this.typeBuilder.getEnumType(this.typeNames, immutable_1.OrderedSet(enumCases));
        }
        else {
            return this.typeBuilder.getStringType(this.typeNames, immutable_1.OrderedMap(counts));
        }
    };
    UnifyUnionBuilder.prototype.makeClass = function (classes, maps) {
        var _this = this;
        if (classes.length > 0 && maps.length > 0) {
            return Support_1.panic("Cannot handle a class type that's also a map");
        }
        if (maps.length > 0) {
            return this.typeBuilder.getMapType(this._unifyTypes(maps, this.typeNames));
        }
        if (classes.length === 1) {
            return this.typeBuilder.lookupTypeRef(classes[0]);
        }
        var maybeTypeRef = this.typeBuilder.lookupTypeRefs(classes);
        if (maybeTypeRef !== undefined) {
            return maybeTypeRef;
        }
        var actualClasses = classes.map(function (c) { return Type_1.assertIsClass(c.deref()[0]); });
        var properties = getCliqueProperties(actualClasses, function (names, types, isNullable) {
            if (types.size === 0) {
                Support_1.assert(isNullable, "Property has no type");
                return _this.typeBuilder.getPrimitiveType("null");
            }
            var tref = _this._unifyTypes(types.map(function (t) { return t.typeRef; }).toArray(), names);
            if (isNullable) {
                return _this.typeBuilder.makeNullable(tref, names);
            }
            return tref;
        });
        return this.typeBuilder.getUniqueClassType(this.typeNames, properties);
    };
    UnifyUnionBuilder.prototype.makeArray = function (arrays) {
        return this.typeBuilder.getArrayType(this._unifyTypes(arrays, this.typeNames.singularize()));
    };
    return UnifyUnionBuilder;
}(TypeBuilder_1.UnionBuilder));
function unifyTypes(types, typeNames, typeBuilder, makeEnums) {
    if (types.isEmpty()) {
        return Support_1.panic("Cannot unify empty set of types");
    }
    else if (types.count() === 1) {
        return typeBuilder.lookupTypeRef(Support_1.defined(types.first()).typeRef);
    }
    var maybeTypeRef = typeBuilder.lookupTypeRefs(types.toArray().map(function (t) { return t.typeRef; }));
    if (maybeTypeRef !== undefined) {
        return maybeTypeRef;
    }
    var unionBuilder = new UnifyUnionBuilder(typeBuilder, typeNames, makeEnums, function (trefs, names) {
        return unifyTypes(immutable_1.Set(trefs.map(function (tref) { return tref.deref()[0]; })), names, typeBuilder, makeEnums);
    });
    var addType = function (t) {
        Type_1.matchTypeExhaustive(t, function (_anyType) { return unionBuilder.addAny(); }, function (_nullType) { return unionBuilder.addNull(); }, function (_boolType) { return unionBuilder.addBool(); }, function (_integerType) { return unionBuilder.addInteger(); }, function (_doubleType) { return unionBuilder.addDouble(); }, function (stringType) {
            var enumCases = stringType.enumCases;
            if (enumCases === undefined) {
                unionBuilder.addStringType("string");
            }
            else {
                unionBuilder.addEnumCases(enumCases);
            }
        }, function (arrayType) { return unionBuilder.addArray(arrayType.items.typeRef); }, function (classType) { return unionBuilder.addClass(classType.typeRef); }, function (mapType) { return unionBuilder.addMap(mapType.values.typeRef); }, 
        // FIXME: We're not carrying counts, so this is not correct if we do enum
        // inference.  JSON Schema input uses this case, however, without enum
        // inference, which is fine, but still a bit ugly.
        function (enumType) { return enumType.cases.forEach(function (s) { return unionBuilder.addEnumCase(s); }); }, function (unionType) { return unionType.members.forEach(addType); }, function (_dateType) { return unionBuilder.addStringType("date"); }, function (_timeType) { return unionBuilder.addStringType("time"); }, function (_dateTimeType) { return unionBuilder.addStringType("date-time"); });
    };
    types.forEach(addType);
    return unionBuilder.buildUnion(false);
}
exports.unifyTypes = unifyTypes;
