#!/usr/bin/env node
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var fs = require("fs");
var path = require("path");
var _ = require("lodash");
// The typings for this module are screwy
var isURL = require("is-url");
var _1 = require(".");
var targetLanguages = require("./Language/All");
var URLGrammar_1 = require("./URLGrammar");
var Annotation_1 = require("./Annotation");
var Support_1 = require("./Support");
var GraphQLIntrospection_1 = require("./GraphQLIntrospection");
var index_1 = require("./get-stream/index");
var commandLineArgs = require("command-line-args");
var getUsage = require("command-line-usage");
var chalk = require("chalk");
var fetch = require("node-fetch");
var packageJSON = require("../package.json");
var langs = targetLanguages.all.map(function (r) { return _.minBy(r.names, function (s) { return s.length; }); }).join("|");
var langDisplayNames = targetLanguages.all.map(function (r) { return r.displayName; }).join(", ");
function readableFromFileOrUrl(fileOrUrl) {
    return __awaiter(this, void 0, void 0, function () {
        var response;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (!isURL(fileOrUrl)) return [3 /*break*/, 2];
                    return [4 /*yield*/, fetch(fileOrUrl)];
                case 1:
                    response = _a.sent();
                    return [2 /*return*/, response.body];
                case 2:
                    if (fs.existsSync(fileOrUrl)) {
                        return [2 /*return*/, fs.createReadStream(fileOrUrl)];
                    }
                    else {
                        return [2 /*return*/, Support_1.panic("Input file " + fileOrUrl + " does not exist")];
                    }
                    _a.label = 3;
                case 3: return [2 /*return*/];
            }
        });
    });
}
function sourceFromFileOrUrlArray(name, filesOrUrls) {
    return __awaiter(this, void 0, void 0, function () {
        var samples;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, Promise.all(filesOrUrls.map(readableFromFileOrUrl))];
                case 1:
                    samples = _a.sent();
                    return [2 /*return*/, { name: name, samples: samples }];
            }
        });
    });
}
function typeNameFromFilename(filename) {
    var name = path.basename(filename);
    return name.substr(0, name.lastIndexOf("."));
}
function samplesFromDirectory(dataDir) {
    return __awaiter(this, void 0, void 0, function () {
        function readFilesOrURLsInDirectory(d) {
            return __awaiter(this, void 0, void 0, function () {
                var files, sourcesInDir, graphQLSources, graphQLSchema, _i, files_1, file, name, fileOrUrl, _a, _b, _c, _d, _e, _f, _g, _h, _j, schema, _k, _l, _m, graphQLSources_1, source;
                return __generator(this, function (_o) {
                    switch (_o.label) {
                        case 0:
                            files = fs
                                .readdirSync(d)
                                .map(function (x) { return path.join(d, x); })
                                .filter(function (x) { return fs.lstatSync(x).isFile(); });
                            sourcesInDir = [];
                            graphQLSources = [];
                            graphQLSchema = undefined;
                            _i = 0, files_1 = files;
                            _o.label = 1;
                        case 1:
                            if (!(_i < files_1.length)) return [3 /*break*/, 10];
                            file = files_1[_i];
                            name = typeNameFromFilename(file);
                            fileOrUrl = file;
                            file = file.toLowerCase();
                            // If file is a URL string, download it
                            if (file.endsWith(".url")) {
                                fileOrUrl = fs.readFileSync(file, "utf8").trim();
                            }
                            if (!(file.endsWith(".url") || file.endsWith(".json"))) return [3 /*break*/, 3];
                            _b = (_a = sourcesInDir).push;
                            _c = {
                                name: name
                            };
                            return [4 /*yield*/, readableFromFileOrUrl(fileOrUrl)];
                        case 2:
                            _b.apply(_a, [(_c.samples = [_o.sent()],
                                    _c)]);
                            return [3 /*break*/, 9];
                        case 3:
                            if (!file.endsWith(".schema")) return [3 /*break*/, 5];
                            _e = (_d = sourcesInDir).push;
                            _f = {
                                name: name
                            };
                            return [4 /*yield*/, readableFromFileOrUrl(fileOrUrl)];
                        case 4:
                            _e.apply(_d, [(_f.schema = _o.sent(),
                                    _f)]);
                            return [3 /*break*/, 9];
                        case 5:
                            if (!file.endsWith(".gqlschema")) return [3 /*break*/, 7];
                            Support_1.assert(graphQLSchema === undefined, "More than one GraphQL schema in " + dataDir);
                            return [4 /*yield*/, readableFromFileOrUrl(fileOrUrl)];
                        case 6:
                            graphQLSchema = _o.sent();
                            return [3 /*break*/, 9];
                        case 7:
                            if (!file.endsWith(".graphql")) return [3 /*break*/, 9];
                            _h = (_g = graphQLSources).push;
                            _j = { name: name, schema: undefined };
                            return [4 /*yield*/, readableFromFileOrUrl(fileOrUrl)];
                        case 8:
                            _h.apply(_g, [(_j.query = _o.sent(), _j)]);
                            _o.label = 9;
                        case 9:
                            _i++;
                            return [3 /*break*/, 1];
                        case 10:
                            if (!(graphQLSources.length > 0)) return [3 /*break*/, 12];
                            if (graphQLSchema === undefined) {
                                return [2 /*return*/, Support_1.panic("No GraphQL schema in " + dataDir)];
                            }
                            _l = (_k = JSON).parse;
                            return [4 /*yield*/, index_1.getStream(graphQLSchema)];
                        case 11:
                            schema = _l.apply(_k, [_o.sent()]);
                            for (_m = 0, graphQLSources_1 = graphQLSources; _m < graphQLSources_1.length; _m++) {
                                source = graphQLSources_1[_m];
                                source.schema = schema;
                                sourcesInDir.push(source);
                            }
                            _o.label = 12;
                        case 12: return [2 /*return*/, sourcesInDir];
                    }
                });
            });
        }
        var contents, directories, sources, _i, directories_1, dir, jsonSamples, schemaSources, graphQLSources, _a, _b, source;
        return __generator(this, function (_c) {
            switch (_c.label) {
                case 0:
                    contents = fs.readdirSync(dataDir).map(function (x) { return path.join(dataDir, x); });
                    directories = contents.filter(function (x) { return fs.lstatSync(x).isDirectory(); });
                    return [4 /*yield*/, readFilesOrURLsInDirectory(dataDir)];
                case 1:
                    sources = _c.sent();
                    _i = 0, directories_1 = directories;
                    _c.label = 2;
                case 2:
                    if (!(_i < directories_1.length)) return [3 /*break*/, 8];
                    dir = directories_1[_i];
                    jsonSamples = [];
                    schemaSources = [];
                    graphQLSources = [];
                    _a = 0;
                    return [4 /*yield*/, readFilesOrURLsInDirectory(dir)];
                case 3:
                    _b = _c.sent();
                    _c.label = 4;
                case 4:
                    if (!(_a < _b.length)) return [3 /*break*/, 6];
                    source = _b[_a];
                    // FIXME: We do a type switch here, but we know which types we're putting in
                    // in the function above.  It should separate it right away.
                    if (_1.isJSONSource(source)) {
                        jsonSamples = jsonSamples.concat(source.samples);
                    }
                    else if (_1.isSchemaSource(source)) {
                        schemaSources.push(source);
                    }
                    else {
                        graphQLSources.push(source);
                    }
                    _c.label = 5;
                case 5:
                    _a++;
                    return [3 /*break*/, 4];
                case 6:
                    if (jsonSamples.length > 0 && schemaSources.length + graphQLSources.length > 0) {
                        return [2 /*return*/, Support_1.panic("Cannot mix JSON samples with JSON Schema or GraphQL in input subdirectory")];
                    }
                    if (schemaSources.length > 0 && graphQLSources.length > 0) {
                        return [2 /*return*/, Support_1.panic("Cannot mix JSON Schema with GraphQL in an input subdirectory")];
                    }
                    if (jsonSamples.length > 0) {
                        sources.push({
                            name: path.basename(dir),
                            samples: jsonSamples
                        });
                    }
                    sources = sources.concat(schemaSources);
                    sources = sources.concat(graphQLSources);
                    _c.label = 7;
                case 7:
                    _i++;
                    return [3 /*break*/, 2];
                case 8: return [2 /*return*/, sources];
            }
        });
    });
}
function inferLang(options) {
    // Output file extension determines the language if language is undefined
    if (options.out) {
        var extension = path.extname(options.out);
        if (extension === "") {
            throw new Error("Please specify a language (--lang) or an output file extension.");
        }
        return extension.substr(1);
    }
    return "go";
}
function inferTopLevel(options) {
    // Output file name determines the top-level if undefined
    if (options.out) {
        var extension = path.extname(options.out);
        var without = path.basename(options.out).replace(extension, "");
        return without;
    }
    // Source determines the top-level if undefined
    if (options.src && options.src.length === 1) {
        var src = options.src[0];
        var extension = path.extname(src);
        var without = path.basename(src).replace(extension, "");
        return without;
    }
    return "TopLevel";
}
function inferOptions(opts) {
    var srcLang = opts.srcLang;
    if (opts.graphqlSchema !== undefined || opts.graphqlIntrospect !== undefined) {
        Support_1.assert(srcLang === undefined || srcLang === "graphql", "If a GraphQL schema is specified, the source language must be GraphQL");
        srcLang = "graphql";
    }
    else {
        Support_1.assert(srcLang !== "graphql", "Please specify a GraphQL schema with --graphql-schema or --graphql-introspect");
        srcLang = srcLang || "json";
    }
    return {
        src: opts.src || [],
        srcLang: srcLang,
        lang: opts.lang || inferLang(opts),
        topLevel: opts.topLevel || inferTopLevel(opts),
        noMaps: !!opts.noMaps,
        noEnums: !!opts.noEnums,
        noCombineClasses: !!opts.noCombineClasses,
        noRender: !!opts.noRender,
        alphabetizeProperties: !!opts.alphabetizeProperties,
        rendererOptions: opts.rendererOptions || {},
        help: opts.help || false,
        quiet: opts.quiet || false,
        version: opts.version || false,
        out: opts.out,
        graphqlSchema: opts.graphqlSchema,
        graphqlIntrospect: opts.graphqlIntrospect,
        graphqlServerHeader: opts.graphqlServerHeader
    };
}
var optionDefinitions = [
    {
        name: "out",
        alias: "o",
        type: String,
        typeLabel: "FILE",
        description: "The output file. Determines --lang and --top-level."
    },
    {
        name: "top-level",
        alias: "t",
        type: String,
        typeLabel: "NAME",
        description: "The name for the top level type."
    },
    {
        name: "lang",
        alias: "l",
        type: String,
        typeLabel: langs,
        description: "The target language."
    },
    {
        name: "src-lang",
        alias: "s",
        type: String,
        defaultValue: undefined,
        typeLabel: "json|schema|graphql",
        description: "The source language (default is json)."
    },
    {
        name: "src",
        type: String,
        multiple: true,
        defaultOption: true,
        typeLabel: "FILE|URL|DIRECTORY",
        description: "The file, url, or data directory to type."
    },
    {
        name: "src-urls",
        type: String,
        typeLabel: "FILE",
        description: "Tracery grammar describing URLs to crawl."
    },
    {
        name: "no-combine-classes",
        type: Boolean,
        description: "Don't combine similar classes."
    },
    {
        name: "graphql-schema",
        type: String,
        typeLabel: "FILE",
        description: "GraphQL introspection file."
    },
    {
        name: "graphql-introspect",
        type: String,
        typeLabel: "URL",
        description: "Introspect GraphQL schema from a server."
    },
    {
        name: "graphql-server-header",
        type: String,
        multiple: true,
        description: "HTTP header for the GraphQL introspection query."
    },
    {
        name: "no-maps",
        type: Boolean,
        description: "Don't infer maps, always use classes."
    },
    {
        name: "no-enums",
        type: Boolean,
        description: "Don't infer enums, always use strings."
    },
    {
        name: "no-render",
        type: Boolean,
        description: "Don't render output."
    },
    {
        name: "alphabetize-properties",
        type: Boolean,
        description: "Alphabetize order of class properties."
    },
    {
        name: "quiet",
        type: Boolean,
        description: "Don't show issues in the generated code."
    },
    {
        name: "help",
        alias: "h",
        type: Boolean,
        description: "Get some help."
    },
    {
        name: "version",
        alias: "v",
        type: Boolean,
        description: "Display the version of quicktype"
    }
];
var sectionsBeforeRenderers = [
    {
        header: "Synopsis",
        content: "$ quicktype [[bold]{--lang} " + langs + "] FILE|URL ..."
    },
    {
        header: "Description",
        content: "Given JSON sample data, quicktype outputs code for working with that data in " + langDisplayNames + "."
    },
    {
        header: "Options",
        optionList: optionDefinitions,
        hide: ["no-render"]
    }
];
var sectionsAfterRenderers = [
    {
        header: "Examples",
        content: [
            chalk.dim("Generate C# to parse a Bitcoin API"),
            "$ quicktype -o LatestBlock.cs https://blockchain.info/latestblock",
            "",
            chalk.dim("Generate Go code from a directory of samples containing:"),
            chalk.dim("  - Foo.json\n  + Bar\n    - bar-sample-1.json\n    - bar-sample-2.json\n  - Baz.url"),
            "$ quicktype -l go samples",
            "",
            chalk.dim("Generate JSON Schema, then TypeScript"),
            "$ quicktype -o schema.json https://blockchain.info/latestblock",
            "$ quicktype -o bitcoin.ts --src-lang schema schema.json"
        ]
    },
    {
        content: "Learn more at [bold]{quicktype.io}"
    }
];
function getOptionDefinitions(opts) {
    return _1.getTargetLanguage(opts.lang).optionDefinitions;
}
function parseArgv(argv) {
    // We can only fully parse the options once we know which renderer is selected,
    // because there are renderer-specific options.  But we only know which renderer
    // is selected after we've parsed the options.  Hence, we parse the options
    // twice.  This is the first parse to get the renderer:
    var incompleteOptions = parseOptions(optionDefinitions, argv, true);
    var rendererOptionDefinitions = getOptionDefinitions(incompleteOptions);
    // Use the global options as well as the renderer options from now on:
    var allOptionDefinitions = _.concat(optionDefinitions, rendererOptionDefinitions);
    try {
        // This is the parse that counts:
        return parseOptions(allOptionDefinitions, argv, false);
    }
    catch (error) {
        if (error.name === "UNKNOWN_OPTION") {
            usage();
            throw new Error("Unknown option");
        }
        throw error;
    }
}
// Parse the options in argv and split them into global options and renderer options,
// according to each option definition's `renderer` field.  If `partial` is false this
// will throw if it encounters an unknown option.
function parseOptions(definitions, argv, partial) {
    var opts = commandLineArgs(definitions, { argv: argv, partial: partial });
    var options = { rendererOptions: {} };
    definitions.forEach(function (o) {
        if (!(o.name in opts))
            return;
        var v = opts[o.name];
        if (o.renderer)
            options.rendererOptions[o.name] = v;
        else {
            var k = _.lowerFirst(o.name
                .split("-")
                .map(_.upperFirst)
                .join(""));
            options[k] = v;
        }
    });
    return inferOptions(options);
}
function usage() {
    var rendererSections = [];
    _.forEach(targetLanguages.all, function (language) {
        var definitions = language.optionDefinitions;
        if (definitions.length === 0)
            return;
        rendererSections.push({
            header: "Options for " + language.displayName,
            optionList: definitions
        });
    });
    var sections = _.concat(sectionsBeforeRenderers, rendererSections, sectionsAfterRenderers);
    console.log(getUsage(sections));
}
function splitAndWriteJava(dir, str) {
    var lines = str.split("\n");
    var filename = null;
    var currentFileContents = "";
    var writeFile = function () {
        if (filename != null) {
            fs.writeFileSync(path.join(dir, filename), currentFileContents);
        }
        filename = null;
        currentFileContents = "";
    };
    var i = 0;
    while (i < lines.length) {
        var line = lines[i];
        i += 1;
        var results = line.match("^// (.+\\.java)$");
        if (results == null) {
            currentFileContents += line + "\n";
        }
        else {
            writeFile();
            filename = results[1];
            while (lines[i] === "")
                i++;
        }
    }
    writeFile();
}
function getSources(options) {
    return __awaiter(this, void 0, void 0, function () {
        var json, jsonMap_1, topLevels, exists, directories_2, sources, _i, directories_3, dataDir, _a, _b, filesOrUrls, _c, _d;
        return __generator(this, function (_e) {
            switch (_e.label) {
                case 0:
                    if (!options.srcUrls) return [3 /*break*/, 1];
                    json = JSON.parse(fs.readFileSync(options.srcUrls, "utf8"));
                    jsonMap_1 = URLGrammar_1.urlsFromURLGrammar(json);
                    topLevels = Object.getOwnPropertyNames(jsonMap_1);
                    return [2 /*return*/, Promise.all(topLevels.map(function (name) { return sourceFromFileOrUrlArray(name, jsonMap_1[name]); }))];
                case 1:
                    if (!(options.src.length === 0)) return [3 /*break*/, 2];
                    return [2 /*return*/, [
                            {
                                name: options.topLevel,
                                samples: [process.stdin]
                            }
                        ]];
                case 2:
                    exists = options.src.filter(fs.existsSync);
                    directories_2 = exists.filter(function (x) { return fs.lstatSync(x).isDirectory(); });
                    sources = [];
                    _i = 0, directories_3 = directories_2;
                    _e.label = 3;
                case 3:
                    if (!(_i < directories_3.length)) return [3 /*break*/, 6];
                    dataDir = directories_3[_i];
                    _b = (_a = sources).concat;
                    return [4 /*yield*/, samplesFromDirectory(dataDir)];
                case 4:
                    sources = _b.apply(_a, [_e.sent()]);
                    _e.label = 5;
                case 5:
                    _i++;
                    return [3 /*break*/, 3];
                case 6:
                    filesOrUrls = options.src.filter(function (x) { return !_.includes(directories_2, x); });
                    if (!!_.isEmpty(filesOrUrls)) return [3 /*break*/, 8];
                    _d = (_c = sources).push;
                    return [4 /*yield*/, sourceFromFileOrUrlArray(options.topLevel, filesOrUrls)];
                case 7:
                    _d.apply(_c, [_e.sent()]);
                    _e.label = 8;
                case 8: return [2 /*return*/, sources];
            }
        });
    });
}
function main(args) {
    return __awaiter(this, void 0, void 0, function () {
        var options, sources, _a, schemaString, wroteSchemaToFile, numSources, gqlSources, _i, _b, queryFile, schemaFile, schema, query, name, _c, _d, source, run, _e, lines_1, annotations, output;
        return __generator(this, function (_f) {
            switch (_f.label) {
                case 0:
                    if (!(_.isArray(args) && args.length === 0)) return [3 /*break*/, 1];
                    usage();
                    return [3 /*break*/, 19];
                case 1:
                    options = _.isArray(args) ? parseArgv(args) : inferOptions(args);
                    if (options.help) {
                        usage();
                        return [2 /*return*/];
                    }
                    if (options.version) {
                        console.log("quicktype, version " + packageJSON.version);
                        console.log("Visit https://quicktype.io/ for more information.");
                        return [2 /*return*/];
                    }
                    sources = [];
                    _a = options.srcLang;
                    switch (_a) {
                        case "graphql": return [3 /*break*/, 2];
                        case "json": return [3 /*break*/, 9];
                        case "schema": return [3 /*break*/, 11];
                    }
                    return [3 /*break*/, 16];
                case 2:
                    schemaString = undefined;
                    wroteSchemaToFile = false;
                    if (!(options.graphqlIntrospect !== undefined)) return [3 /*break*/, 4];
                    return [4 /*yield*/, GraphQLIntrospection_1.introspectServer(options.graphqlIntrospect, options.graphqlServerHeader || [])];
                case 3:
                    schemaString = _f.sent();
                    if (options.graphqlSchema !== undefined) {
                        fs.writeFileSync(options.graphqlSchema, schemaString);
                        wroteSchemaToFile = true;
                    }
                    _f.label = 4;
                case 4:
                    numSources = options.src.length;
                    if (numSources !== 1) {
                        if (wroteSchemaToFile) {
                            // We're done.
                            return [2 /*return*/];
                        }
                        if (numSources === 0) {
                            return [2 /*return*/, Support_1.panic("Please specify at least one GraphQL query as input")];
                        }
                    }
                    gqlSources = [];
                    _i = 0, _b = options.src;
                    _f.label = 5;
                case 5:
                    if (!(_i < _b.length)) return [3 /*break*/, 8];
                    queryFile = _b[_i];
                    if (schemaString === undefined) {
                        schemaFile = Support_1.defined(options.graphqlSchema);
                        schemaString = fs.readFileSync(schemaFile, "utf8");
                    }
                    schema = JSON.parse(schemaString);
                    return [4 /*yield*/, readableFromFileOrUrl(queryFile)];
                case 6:
                    query = _f.sent();
                    name = numSources === 1 ? options.topLevel : typeNameFromFilename(queryFile);
                    gqlSources.push({ name: name, schema: schema, query: query });
                    _f.label = 7;
                case 7:
                    _i++;
                    return [3 /*break*/, 5];
                case 8:
                    sources = gqlSources;
                    return [3 /*break*/, 17];
                case 9: return [4 /*yield*/, getSources(options)];
                case 10:
                    sources = _f.sent();
                    return [3 /*break*/, 17];
                case 11:
                    _c = 0;
                    return [4 /*yield*/, getSources(options)];
                case 12:
                    _d = _f.sent();
                    _f.label = 13;
                case 13:
                    if (!(_c < _d.length)) return [3 /*break*/, 15];
                    source = _d[_c];
                    if (_1.isGraphQLSource(source)) {
                        return [2 /*return*/, Support_1.panic("Cannot accept GraphQL for JSON Schema input")];
                    }
                    if (_1.isJSONSource(source)) {
                        Support_1.assert(source.samples.length === 1, "Please specify one schema file for " + source.name);
                        sources.push({
                            name: source.name,
                            schema: source.samples[0]
                        });
                    }
                    else {
                        sources.push(source);
                    }
                    _f.label = 14;
                case 14:
                    _c++;
                    return [3 /*break*/, 13];
                case 15: return [3 /*break*/, 17];
                case 16:
                    Support_1.panic("Unsupported source language (" + options.srcLang + ")");
                    return [3 /*break*/, 17];
                case 17:
                    run = new _1.Run({
                        lang: options.lang,
                        sources: sources,
                        inferMaps: !options.noMaps,
                        inferEnums: !options.noEnums,
                        alphabetizeProperties: options.alphabetizeProperties,
                        combineClasses: !options.noCombineClasses,
                        noRender: options.noRender,
                        rendererOptions: options.rendererOptions
                    });
                    return [4 /*yield*/, run.run()];
                case 18:
                    _e = _f.sent(), lines_1 = _e.lines, annotations = _e.annotations;
                    output = lines_1.join("\n");
                    if (options.out) {
                        if (options.lang === "java") {
                            splitAndWriteJava(path.dirname(options.out), output);
                        }
                        else {
                            fs.writeFileSync(options.out, output);
                        }
                    }
                    else {
                        process.stdout.write(output);
                    }
                    if (options.quiet) {
                        return [2 /*return*/];
                    }
                    annotations.forEach(function (sa) {
                        var annotation = sa.annotation;
                        if (!(annotation instanceof Annotation_1.IssueAnnotationData))
                            return;
                        var lineNumber = sa.span.start.line;
                        var humanLineNumber = lineNumber + 1;
                        console.error("\nIssue in line " + humanLineNumber + ": " + annotation.message);
                        console.error(humanLineNumber + ": " + lines_1[lineNumber]);
                    });
                    _f.label = 19;
                case 19: return [2 /*return*/];
            }
        });
    });
}
exports.main = main;
if (require.main === module) {
    main(process.argv.slice(2)).catch(function (e) {
        if (e instanceof Error) {
            console.error("Error: " + e.message + ".");
        }
        else {
            console.error(e);
        }
        process.exit(1);
    });
}
