/// <reference types="node" />
import { Readable } from "stream";
import { TargetLanguage } from "./TargetLanguage";
import { SerializedRenderResult } from "./Source";
export { TargetLanguage } from "./TargetLanguage";
export { SerializedRenderResult, Annotation } from "./Source";
export { all as languages, languageNamed } from "./Language/All";
export { OptionDefinition } from "./RendererOptions";
export declare function getTargetLanguage(name: string): TargetLanguage;
export declare type RendererOptions = {
    [name: string]: string;
};
export declare type StringInput = string | Readable;
export interface JSONTypeSource {
    name: string;
    samples: StringInput[];
}
export declare function isJSONSource(source: TypeSource): source is JSONTypeSource;
export interface SchemaTypeSource {
    name: string;
    schema: StringInput;
}
export declare function isSchemaSource(source: TypeSource): source is SchemaTypeSource;
export interface GraphQLTypeSource {
    name: string;
    schema: any;
    query: StringInput;
}
export declare function isGraphQLSource(source: TypeSource): source is GraphQLTypeSource;
export declare type TypeSource = GraphQLTypeSource | JSONTypeSource | SchemaTypeSource;
export interface Options {
    lang: string;
    sources: TypeSource[];
    inferMaps: boolean;
    inferEnums: boolean;
    alphabetizeProperties: boolean;
    combineClasses: boolean;
    noRender: boolean;
    leadingComments: string[] | undefined;
    rendererOptions: RendererOptions;
    indentation: string | undefined;
}
export declare class Run {
    private _compressedJSON;
    private _allInputs;
    private _options;
    constructor(options: Partial<Options>);
    private makeGraph;
    run: () => Promise<SerializedRenderResult>;
}
export declare function quicktype(options: Partial<Options>): Promise<SerializedRenderResult>;
